/*
  xobject.c

  Copyright (c) 1995-96 by Dragos Acostachioaie & Adrian Teohar

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <ncurses.h>
#include <signal.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/kd.h>

#include "tscreen.h"
#include "xobject.h"

unsigned ispcconsole, isconsole, havecolors, cursorvisible;

XObject::XObject(XRect _bounds)
{
  state = 0;
  server = 0;
  client = 0;
  objclass = OC_XOBJECT;
  insertObject(this);
  origin = _bounds.a - XPoint(1, 1);
  size = _bounds.b - _bounds.a + XPoint(1, 1);
  _bounds.a -= XPoint(1, 1);
  _bounds.b -= XPoint(1, 1);
  bounds = _bounds;
}

inline XObject::~XObject()
{
  deleteObject(id);
  if(client)
    delete client;
}

inline void XObject::calculatebounds()
{
  if(server)
    {
    bounds.a = server->bounds.a + origin;
    bounds.b = bounds.a + size;
    }
  else
    {
    bounds.a = origin;
    bounds.b = origin + size;
    }
}

inline void XObject::getevent(XEvent *&ev)
{
  mailer->getevent(ev, id);
}

inline int XObject::getstate(int mask)
{
  return(state & mask);
}

void XObject::registerclient(XObject *&newclient)
/* this will register the new client and will adjust the client's bounds */
{
  XObject *temp = client;
  /* "client" is a linked list, so we insert at the end of list */
  if(client == 0)
    client = newclient;
  else
    {
    while(temp->client)
      temp = temp->client;
    temp->client = newclient;
    newclient->client = 0;
    }
  newclient->server = this;
  newclient->calculatebounds();
}

inline void XObject::setstate(int mask, int set_bool)
{
  if(state & mask)
    {
    if(set_bool == 0)
      state -= mask;
    }
  else
    if(set_bool)
      state += mask;
}

inline char *XObject::getpalette()
{
  return pal; 
}

inline void XObject::setpalette(char *_pal)
{
  pal = _pal;
  if(getstate(SM_VISIBLE))
    draw();
}

inline void XObject::moveto(XPoint p)
{
  origin = p;
  calculatebounds();
  bounds.b -= XPoint(1, 1);
}

unsigned newid = 0x0f;   /* ids from 0 to 0x0f are reserved */
xobjectTable *objectTable = 0;

void insertObject(XObject *&ptr)
{
  xobjectTable *temp, *newobject;
  ptr->id = newid;
  newid++;
  if(objectTable == 0)
    {
    objectTable = new xobjectTable;
    objectTable->oid = ptr->id;
    objectTable->ptr = ptr;
    objectTable->next = 0;
    }
  else
    {
    temp = objectTable;
    while(temp->next)
      temp = temp->next;
    newobject = new xobjectTable;
    newobject->ptr = ptr;
    newobject->oid = ptr->id;
    newobject->next = 0;
    temp->next = newobject;
    }
}

void deleteObject(unsigned id)
{
  xobjectTable *prev = NULL, *temp = objectTable;
  if(id >= newid)
    return;
  while(temp->next && temp->oid != id)
    {
    prev = temp;
    temp = temp->next;
    }
  if(temp->oid == id)
    {
    if(temp == objectTable)
      {
      objectTable = objectTable->next;
      delete temp;
      }
    else
     {
      prev->next = temp->next;
      delete temp;
     }
    }
}

XObject* getoptr(unsigned id)
{
  xobjectTable *temp = objectTable;
  while(temp->next && temp->oid != id)
    temp = temp->next;
  if(temp->oid == id)
    return temp->ptr;
  return 0;
}

unsigned getoid(XObject *ptr)
{
  xobjectTable *temp = objectTable;
  while(temp->next && temp->ptr != ptr)
    temp = temp->next;
  if(temp->ptr == ptr)
    return temp->oid;
  return 0;
}

int isObjectRegistered(unsigned id)
{
  xobjectTable *temp = objectTable;
  while(temp->next && temp->oid != id)
    temp = temp->next;
  if(temp->oid == id)
    return 1;
  return 0;
}

int isObjectRegistered(XObject *ptr)
{
  xobjectTable *temp = objectTable;
  while(temp->next && temp->ptr != ptr)
    temp = temp->next;
  if(temp->ptr == ptr)
    return 1;
  return 0;
}

static void exitnow(int sig)
{
  xdone();
  exit(sig);
}

void xinit()
{
  isconsole = ioctl(0, KDGKBMETA, &isconsole);
  initscr();
  start_color();
  cbreak();
  noecho();
  keypad(stdscr, 1);
  curs_set(0);
  refresh();
  havecolors = has_colors();
  if(strcmp(termname(), "linux"))
    ispcconsole = 0;
   else
    ispcconsole = 1;
  cursorvisible = 0;
  initEvents();
  signal(SIGINT, exitnow);
  signal(SIGQUIT, exitnow);
  signal(SIGTERM, exitnow);
}

void xdone()
{
  curs_set(1);
  cursorvisible = 1;
  clear();
  refresh();
  endwin();
  doneEvents();
}
