/*
  xtext.c
  
  Copyright (c) 1995-96 by Dragos Acostachioaie & Adrian Teohar

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <ncurses.h>
#include <stdlib.h>
#include <string.h>

#include "bclass.h"
#include "event.h"
#include "tscreen.h"
#include "xobject.h"
#include "xtext.h"

#define wndPalette "\x07\x04\x07\x04\x07\x04\x00\x04\x00\x04\x07\x00"

XtWindow::XtWindow(XRect r, char *_title) :
  XObject(r)
{
  objclass = "XTWINDOW";
  title = _title;
  pal = wndPalette;
  if(ispcconsole)
    frame = FS_DOUBLEFRAME;
/* by default, Xt will choose the best chars for the current terminal */
   else
    {
    frame = fs_default;
    }
  buffer = new chtype[size.y * size.x];
  XObject::setstate(SM_FOCUSED, 1);
}

XtWindow::~XtWindow()
{
  if(getstate(SM_VISIBLE))
    close();
  delete buffer;
}

void XtWindow::draw()
{
  XObject *index = client;
  if(!getstate(SM_VISIBLE))
    getwindow(bounds, buffer);
  XObject::setstate(SM_VISIBLE, 1);
  frameupdate();
  colorwindow(XRect(bounds.a.x + 1, bounds.a.y + 1, bounds.b.x - 1, bounds.b.y - 1), pal[0], pal[1], A_BOLD);
  refresh();
  while(index)
    {
    index->draw();
    index = index->client;
    }
}

void XtWindow::close()
{
  if(getstate(SM_VISIBLE))
    {
    putwindow(bounds, buffer);
    refresh();
    XObject::setstate(SM_VISIBLE, 0);
    }
}

void XtWindow::handleevent(XEvent *&ev)
{
  if(ev->what == EV_MESSAGE)
    switch(ev->evMessage.message)
      {
      case WM_CLOSE:
        close();
        break;
      case WM_OPEN:
        draw();
        break;
      }
}

void XtWindow::frameupdate()
{
  div_t x;
  unsigned xi;
  if(getstate(SM_FOCUSED))
    init_pair(1, pal[2], pal[3]);
   else
    init_pair(1, pal[6], pal[7]);
  attrset(COLOR_PAIR(1));
  _border(bounds, frame);
  if(title)
    {
    if(getstate(SM_FOCUSED))
      init_pair(1, (unsigned char)pal[4], pal[5]);
     else
      init_pair(1, (unsigned char)pal[8], pal[9]);
    attrset(COLOR_PAIR(1));
    x = div(size.x - strlen(title), 2);
    xi = x.quot + bounds.a.x;
    mvaddch(bounds.a.y, xi - 1, ' ');
    mvaddstr(bounds.a.y, xi, title);
    mvaddch(bounds.a.y, xi + strlen(title), ' ');
    }
}

void XtWindow::setstate(int mask, int set_bool)
{
  XObject::setstate(mask, set_bool);
  if(getstate(SM_VISIBLE))
    frameupdate();
}

void XtWindow::setframe(char *_frame)
{
  frame = _frame;
  if(getstate(SM_VISIBLE))
    frameupdate();
}

void XtWindow::moveto(XPoint p)
{
  chtype *temp = new chtype[size.y * size.x];
  XRect old = XRect(bounds.a - XPoint(1, 1), bounds.b + XPoint(1, 1));
  getwindow(bounds, temp);
  putwindow(bounds, buffer);
  XObject::moveto(p);
  getwindow(bounds, buffer);
  putwindow(bounds, temp);
  delete temp;
  for(unsigned i = bounds.a.y; i <= bounds.b.y; i++)
    {
    stdscr->_line[i].firstchar = old.a.x;
    stdscr->_line[i].lastchar = old.b.x;
    }
  refresh();
}

#define bgPalette "\x07\x04"

XtBackground::XtBackground(XRect r, chtype ch)
{
  bounds = r;
  pattern = ch;
  pal = bgPalette;
}

void XtBackground::draw()
{
  init_pair(1, pal[0], pal[0]);
  attrset(COLOR_PAIR(1));
  for(unsigned j = bounds.a.y; j <= bounds.b.y; j++)
    for(unsigned i = bounds.a.x; i <= bounds.b.x; i++)
      mvaddch(j, i, pattern);
  refresh();
}

void XtBackground::setpalette(char *_pal)
{
  pal = _pal;
  draw();
}

void XtBackground::setpattern(chtype _pattern)
{
  pattern = _pattern;
  draw();
}

#define stPalette "\x00\x07"

XtStaticText::XtStaticText(XRect r, unsigned _alignment, char *_text) :
  XObject(r)
{
  objclass = "XTSTATICTEXT";
  text = _text;
  alignment = _alignment;
  pal = stPalette;
}

void XtStaticText::draw()
{
  XPoint pos;
  div_t x, y;
  pos = bounds.a;
  switch(alignment) 
    {
    case AS_LEFT:
      pos.x = server->bounds.a.x + 1;
      break;
    case AS_CENTER:
      x = div(server->bounds.b.x - server->bounds.a.x - strlen(text), 2);
      pos.x = server->bounds.a.x + x.quot;
      break;
    case AS_RIGHT:
      pos.x = server->bounds.b.x - strlen(text);
      break;
    case AS_TOP:
      pos.y = server->bounds.a.y + 1;
      break;
    case AS_BOTTOM:
      pos.y = server->bounds.a.y - 1;
      break;
    case AS_MIDDLE:
      x = div(server->bounds.b.x - server->bounds.a.x - strlen(text), 2);
      pos.x = server->bounds.a.x + x.quot;
      y = div(server->bounds.b.y - server->bounds.a.y, 2);
      pos.y = server->bounds.a.y + y.quot - 1;
      break;
    default:
      pos.x = bounds.a.x;
      pos.y = bounds.a.y;
    }
  init_pair(1, pal[0], pal[1]);
  attrset(COLOR_PAIR(1));
  mvaddstr(pos.y, pos.x, text);
  refresh();
}

void XtStaticText::setvalue(char *_text)
{
  text = _text;
  if(getstate(SM_VISIBLE))
    draw();
}

#define isfPalette "\x07\x00\x00\x06"

#define IS_MAXSTRLEN 255

XtInputStringField::XtInputStringField(XRect r, char *_defaultstring, unsigned _maxlen, unsigned _encrypt) :
  XObject(r)
{
  objclass = "XTINPUTSTRINGFIELD";
  fieldlength = size.x - 2;
  maxlen = _maxlen;
  encrypt = _encrypt;
  pal = isfPalette;
  currstring = new char[IS_MAXSTRLEN];
  if(encrypt)
    {
    defaultstring = 0;
    maxlen = fieldlength + 1;
    stringlength = 0;
    bzero(currstring, IS_MAXSTRLEN);
    bchar = IS_EBACKGROUNDCHAR;
    }
   else
    {
    defaultstring = _defaultstring;
    stringlength = strlen(defaultstring);
    strcpy(currstring, defaultstring);
    bchar = ' ';
    }
  insertmode = 1;
  inputtext = 0;
  CP = 1;
  curX = 1;
  XObject::setstate(SM_ACTIVE, 1);
  if(cursorvisible)
    XObject::setstate(SM_CURSORVISIBLE, 1);
}

XtInputStringField::~XtInputStringField()
{
  delete currstring;
}

void XtInputStringField::draw()
{
  char *temp = new char[stringlength + 1];
  XObject::setstate(SM_VISIBLE, 1);
  cursor = bounds.a;
  cursor.x += curX;
  init_pair(1, pal[0], pal[1]);
  init_pair(2, pal[2], pal[3]);
  attrset(COLOR_PAIR(1));
  fillwindow(bounds, bchar);
  if(encrypt)
    {
    mvaddch(bounds.a.y, bounds.a.x, IS_ELEFTMARGIN);
    mvaddch(bounds.a.y, bounds.b.x, IS_ERIGHTMARGIN);
    for(unsigned i = 1; i <= stringlength; i++)
      mvaddch(bounds.a.y, bounds.a.x + i, IS_ENCRYPTCONTROLCHAR);
    move(cursor.y, cursor.x);
    refresh();
    return;
    }
  if(CP >= fieldlength + 2)
    {
    strcpy(temp, currstring + CP - fieldlength - 1);
    temp[fieldlength + 1] = '\0';
    mvaddch(bounds.a.y, bounds.a.x, IS_LEFTMARGIN);
    }
   else
    {
    strcpy(temp, currstring);
    if(stringlength > fieldlength)
      temp[fieldlength + 1] = '\0';
     else
      temp[stringlength] = '\0';
    }
  if(CP + fieldlength - curX < stringlength)
    mvaddch(bounds.a.y, bounds.b.x, IS_RIGHTMARGIN);
  if(getstate(SM_FOCUSED))
    if(!inputtext)
      attrset(COLOR_PAIR(2));
  mvaddstr(bounds.a.y, bounds.a.x + 1, temp);
  move(cursor.y, cursor.x);
  delete temp;
  refresh();
}

void XtInputStringField::handleevent(XEvent *&ev)
{
  unsigned key, oldtext;
  switch(ev->what)
    {
    case EV_NOTHING:
      return;
    case EV_MESSAGE:
      switch(ev->evMessage.message)
        {
        case MT_RELEASEFOCUS:
          curs_set(0);
          setstate(SM_FOCUSED, 0);
          return;
        case MT_RESUMEFOCUS:
          if(insertmode)
            curs_set(1);
           else
            curs_set(2);
          setstate(SM_FOCUSED, 1);
          return;
        }
      break;
    case EV_KEYDOWN:
      oldtext = inputtext;
      inputtext = 1;
      switch(ev->evKeyboard.keycode)
        {
        case KEY_IC:
          if(!encrypt)
            {
            insertmode = !insertmode;
            if(insertmode)
              curs_set(1);
             else
              curs_set(2);
            draw();
            }
          break;
        case 0x0d:
          break;
        case 0x1b:
          break;
        case KEY_LEFT:
          if(CP > 1 && !encrypt)
            {
            CP--;
            if(curX > 1 && CP < fieldlength + 1)
              curX--;
            draw();
            }
          break;
        case KEY_RIGHT:
          if(CP <= stringlength && !encrypt)
            {
            CP++;
            if(curX <= fieldlength)
              curX++;
            draw();
            }
          break;
        case KEY_HOME:
          if(CP > 1 && !encrypt)
            {
            CP = 1;
            curX = 1;
            draw();
            }
          break;
        case KEY_END:
          if(CP <= stringlength && !encrypt)
            {
            CP = stringlength + 1;
            if(stringlength > fieldlength)
              curX = fieldlength + 1;
             else
              curX = stringlength + 1;
            draw();
            }
          break;
        case KEY_BACKSPACE:
        case KB_BACK:
          if(CP > 1)
            {
            if(CP == stringlength + 1)
              currstring[stringlength - 1] = '\0';
             else
              {
              for(unsigned i = CP - 1; i <= stringlength - 1; i++)
                currstring[i - 1] = currstring[i];
              currstring[stringlength - 1] = '\0';
              }
            CP--;
            if(CP < fieldlength + 1 || CP != stringlength)
              curX--;
            stringlength--;
            draw();
            }
          break;
        case KEY_DC:
          if(CP <= stringlength)
            {
            for(unsigned i = CP; i <= stringlength - 1; i++)
              currstring[i - 1] = currstring[i];
            currstring[stringlength - 1] = '\0';
            stringlength--;
            draw();
            }
          break;
        default:
          key = ev->evKeyboard.keycode;
          if(key >= 0x20 && key <= 0xff && stringlength < maxlen)
            {
            if(!oldtext)
              {
              stringlength = 0;
              bzero(currstring, IS_MAXSTRLEN);
              }
            if(CP <= stringlength)
              if(insertmode)
                {
                for(unsigned i = stringlength; i >= CP; i--)
                  currstring[i] = currstring[i - 1];
                currstring[CP - 1] = key;
                currstring[stringlength + 1] = '\0';
                stringlength++;
                }
               else
                currstring[CP - 1] = key;
             else
              {
              currstring[stringlength] = key;
              currstring[stringlength + 1] = '\0';
              stringlength++;
              }
            CP++;
            if(curX <= fieldlength)
              curX++;
            draw();
            }
          break;
        }
      break;
    }
}

void XtInputStringField::setstate(int mask, int set_bool)
{
  XObject::setstate(mask, set_bool);
  if(getstate(SM_VISIBLE))
    draw();
}

#define sbPalette "\x07\x06\x06\x07"

XtScrollBar::XtScrollBar(XRect r, long totalno, bool _horizontal) :
  XObject(r)
{
  currentno = 0;
  maxno = totalno;
  horizontal = _horizontal;
  pal = sbPalette;
}

void XtScrollBar::draw()
{
  unsigned i, poz;
  setstate(SM_VISIBLE, 1);
  init_pair(1, pal[0], pal[1]);
  init_pair(2, pal[2], pal[3]);
  attrset(COLOR_PAIR(2) | A_BOLD);
  mvaddch(bounds.a.y, bounds.a.x, ACS_LARROW);
  if(horizontal)
    {
    poz = ((size.x - 2) * currentno) / maxno + 1;
    mvaddch(bounds.a.y, bounds.b.x, ACS_RARROW);
    attrset(COLOR_PAIR(1));
    for(i = 1; i < size.x - 1; i++)
      mvaddch(bounds.a.y, bounds.a.x + i, ACS_BOARD);
    attron(A_BOLD);
    mvaddch(bounds.a.y, bounds.a.x + poz, ACS_BLOCK);
    }
  else
    {
    poz = ((size.y - 2) * currentno) / maxno + 1;
    mvaddch(bounds.b.y, bounds.a.x, ACS_RARROW);
    attrset(COLOR_PAIR(1));
    for(i = 1; i < size.y - 1; i++)
      mvaddch(bounds.a.y + i, bounds.a.x, ACS_BOARD);
    attron(A_BOLD);
    mvaddch(bounds.a.y + poz, bounds.a.x, ACS_BLOCK);
    }  
  refresh();
}

void XtScrollBar::setvalue(long itemno)
{
  currentno = itemno;
  if(getstate(SM_VISIBLE))
    draw();
}

void XtScrollBar::settotalno(long itemno)
{
  maxno = itemno;
  if(getstate(SM_VISIBLE))
    draw();
}

long XtScrollBar::gettotalno()
{
  return maxno;
}

XtProgressBar::XtProgressBar(XRect r, long totalno, unsigned _direction) :
  XObject(r)
{
  maxno = totalno;
  currentno = 0;
  percent = 0;
  poz = 0;
  if(direction <= PB_RIGHTTOLEFT)
    len = size.x;
   else
    len = size.y;
  direction = _direction;
  pal = sbPalette;
}

void XtProgressBar::draw()
{
  unsigned i, newpoz;
  setstate(SM_VISIBLE, 1);
  if(currentno >= 0 && currentno <= maxno)
    {
    percent = (currentno * 100) / maxno;
    init_pair(1, pal[2], pal[3]);
    attrset(COLOR_PAIR(1) | A_BOLD);
    newpoz = (len * currentno) / maxno;
    if(poz <= newpoz)
      {
      for(i = poz; i <= newpoz; i++)
        switch(direction) 
          {
          case PB_LEFTTORIGHT:
            mvaddch(bounds.a.y, bounds.a.x + i - 1, ACS_BLOCK);
            break;
          case PB_RIGHTTOLEFT:
            mvaddch(bounds.a.y, bounds.b.x - i + 1, ACS_BLOCK);
            break;
          case PB_DOWNTOUP:
            mvaddch(bounds.b.y - i + 1, bounds.a.x, ACS_BLOCK);
            break;
          case PB_UPTODOWN:
            mvaddch(bounds.a.y + i - 1, bounds.a.x, ACS_BLOCK);
            break;
          }
      init_pair(1, pal[0], pal[1]);
      attrset(COLOR_PAIR(1));
      for(i = newpoz + 1; i < len; i++)
        switch(direction) 
          {
          case PB_LEFTTORIGHT:
            mvaddch(bounds.a.y, bounds.a.x + i - 1, ACS_BOARD);
            break;
          case PB_RIGHTTOLEFT:
            mvaddch(bounds.a.y, bounds.b.x - i + 1, ACS_BOARD);
            break;
          case PB_DOWNTOUP:
            mvaddch(bounds.b.y - i + 1, bounds.a.x, ACS_BOARD);
            break;
          case PB_UPTODOWN:
            mvaddch(bounds.a.y + i - 1, bounds.a.x, ACS_BOARD);
            break;
          }
      }
    refresh();
    }
}

float XtProgressBar::getpercent()
{
  return percent;
}

void XtProgressBar::setvalue(long itemno)
{
  currentno = itemno;
  if(getstate(SM_VISIBLE))
    draw();
}

long XtProgressBar::getvalue()
{
  return currentno;
}

void XtProgressBar::settotalno(long itemno)
{
  maxno = itemno;
  if(getstate(SM_VISIBLE))
    draw();
}

long XtProgressBar::gettotalno()
{
  return maxno;
}
